#pragma once

// C++ system level
#include <set>
#include <cstring>     // memset, strncmp
// #include <cstdlib>     // TODO: is this for memcpy?
// #include <functional>

// ESP32 specific
// #include "esp_err.h"
#include "nvs_flash.h"
#include "esp_log.h"
// #include "driver/gpio.h"
// #include "driver/uart.h"

// project specific
#include <Types.hpp>
// #include "SpiderLib/ManagedMutex.hpp"

// qthing stuff
// #include <qthing>


namespace SiliconTorch {

  namespace NVSExplorer {


    extern const char* TAG;


    class NVSExplorer {

      public:

        std::set<str> listNamespaces();
        std::set<str> listKeys(const str& nameSpace);

        /* TODO: implement error return!! */  bool removeKey(const str& nameSpace, const str& key);


        bool setU8(const  str& nameSpace, const str& key, u8  value);
        bool setI8(const  str& nameSpace, const str& key, i8  value);

        bool setU16(const str& nameSpace, const str& key, u16 value);
        bool setI16(const str& nameSpace, const str& key, i16 value);

        bool setU32(const str& nameSpace, const str& key, u32 value);
        bool setI32(const str& nameSpace, const str& key, i32 value);

        bool setU64(const str& nameSpace, const str& key, u64 value);
        bool setI64(const str& nameSpace, const str& key, i64 value);


        u8  getU8(const  str& nameSpace, const str& key, u8  defaultValue = 0);
        i8  getI8(const  str& nameSpace, const str& key, i8  defaultValue = 0);

        u16 getU16(const str& nameSpace, const str& key, u16 defaultValue = 0);
        i16 getI16(const str& nameSpace, const str& key, i16 defaultValue = 0);

        u32 getU32(const str& nameSpace, const str& key, u32 defaultValue = 0);
        i32 getI32(const str& nameSpace, const str& key, i32 defaultValue = 0);

        u64 getU64(const str& nameSpace, const str& key, u64 defaultValue = 0);
        i64 getI64(const str& nameSpace, const str& key, i64 defaultValue = 0);


        float  getFloat(const str& nameSpace, const str& key);
        double getDouble(const str& nameSpace, const str& key);

        bool setFloat(const str& nameSpace, const str& key, float value);
        bool setDouble(const str& nameSpace, const str& key, double value);

        // Gets the key regardless of its type in flash
        i64  getSignedInt(const str& nameSpace, const str& key, i64 defaultValue = 0);
        u64 getUnsignedInt(const str& nameSpace, const str& key, u64 defaultValue = 0);

        // TODO: ⮦ Both unimplemented atm ⮧ 
        bool setSignedInt(const str& nameSpace, const str& key, i64 value, nvs_type_t entryType = NVS_TYPE_ANY);
        bool setUnsignedInt(const str& nameSpace, const str& key, u64 value, nvs_type_t entryType = NVS_TYPE_ANY);


        size_t getStringLength(const str& nameSpace, const str& key, size_t defaultValue = 0);
        str getString(const str& nameSpace, const str& key, const str& defaultValue = "");
        // size_t getLongString(const str& nameSpace, const str& key, char* const buffer);

        bool setString(const str& nameSpace, const str& key, const str& value);


        /* TODO: implement! */       bool copyKey(const str& nameSpace, const str& keyFrom, const str& keyTo);                // copies key inside namespace
        /* TODO: implement! */       bool copyKey(const str& nsFrom, const str& keyFrom, const str& nsTo, const str& keyTo);  // copies key to another namespace

        /* TODO: test thoroughly! */ bool moveKey(const str& nameSpace, const str& keyFrom, const str& keyTo);                // moves key inside namespace
        /* TODO: test thoroughly! */ bool moveKey(const str& nsFrom, const str& keyFrom, const str& nsTo, const str& keyTo);  // moves key to another namespace


        str type2str(nvs_type_t typ);
        nvs_type_t getType(const str& nameSpace, const str& key);
        str getTypeStr(const str& nameSpace, const str& key);


        str deviceTopic(const str& suffix);

        str requestTopic(const str& cmd);
        str requestTopic(const str& cmd, const str& nameSpace);
        str requestTopic(const str& cmd, const str& nameSpace, const str& key);

        str responseTopic(const str& cmd);
        str responseTopic(const str& cmd, const str& nameSpace);
        str responseTopic(const str& cmd, const str& nameSpace, const str& key);

        // ⬇⬇ Singleton stuff ⬇⬇

        static NVSExplorer& instance() {
          static NVSExplorer _instance;
          return _instance;
        }

        NVSExplorer(NVSExplorer const&)    = delete;
        void operator=(NVSExplorer const&) = delete;

        // ⬆⬆ Singleton stuff ⬆⬆

      private:

        void registerLambdas0();
        void registerLambdas1();

        NVSExplorer();

    };


  }
}
